Require Import ssreflect ssrbool ssrnat eqtype ssrfun seq path.
Require Import pred prelude idynamic ordtype finmap pcm coding unionmap heap.
Require Import feap state zigzag world inject rely hide atomic autosplit.
Require Import privstate footprint model lemmas locks extra_lemmas stsep.
Require Import getter.
Set Implicit Arguments.
Unset Strict Implicit.
Unset Printing Implicit Defensive.

(*** A rough-and-ready implementation of an idealistic coarse-grained allocator ***)

Module DummyAllocator.
Section DummyAllocator.
Variables (hp al : nat) (sent : ptr).

(* A heap predicate, defining a memory pool with a structure,
   described by the purely functional list avl.  *)

Fixpoint alloc_inv' avl := 
  [Pred h | if avl is x::xs 
            then exists h', h = x :-> tt \+ h' /\ alloc_inv' xs h'
            else h = Unit].

Lemma alloc_inv_prec avl h1 h2 :
        alloc_inv' avl h1 -> alloc_inv' avl h2 -> h1 = h2.
Proof. 
elim: avl h1 h2=>[|x xs IH] h1 h2 /=; first by move=>->->.
by case=>h1' [Z1 H1][h2'][Z2 H2]; subst h1 h2; rewrite (IH _ _ H1 H2).
Qed.

(* Allocator invariant, requireing a heap contain the sentinel pointer
  sent, which points to the first element of the memory pool h'. *)
Definition alloc_inv  (g : unit) : Pred heap := 
  [Pred h | exists (avl: seq ptr) h', 
             [/\ valid h, h = sent :-> avl \+ h' & 
                 alloc_inv' avl h']].

Lemma ainv_prec (g : [encoded_pcm of unit]) : precise (@alloc_inv g).
Proof.
case: g=>h1' h2' t1 t2 Vh Eh.
case=>avl1 [h1][Vh1]Z1 Hm1; subst h1'.
case=>avl2 [h2][Vh2]Z2 Hm2; subst h2'.
rewrite -!joinA in Vh Eh.
case:(hcancelV Vh Eh)=>Z1 V1 E1; subst avl2.
by move: (alloc_inv_prec Hm1 Hm2)=>->.
Qed.

(* Simple allocator is parametrised by the Priv-label hp. It also takes
a 'sent' and the 'al' pointers explicitly and returns a rich cas_lock
structure with everything instantiated. *)

Definition alloc_lock_info := LockInfo al ainv_prec. 
Variable Alloc : lockmod alloc_lock_info.
Notation W := (entangle (Priv hp) Alloc). 
Notation locked := (locked Alloc tt tt).
Notation lock := (lock Alloc hp). Notation unlock := (unlock Alloc hp). 
Notation coh := (Mod.coh W). Notation "[ e ]" := (extend _ e).
Notation priv_read := lemmas.priv_read. Notation priv_write := lemmas.priv_write.

Definition F := fun _: pcm_of alloc_lock_info => tt.
Lemma pfF : local F. Proof. by []. Qed.

Lemma is_locked_stable : menv_stable Alloc locked. 
Proof. by apply: locked_stable. Qed.

Lemma coh_is_lock (hS hO: heap) st:
        hp ->> [hS, tt, hO] \+ st \In coh -> 
        is_lock Alloc tt tt st.
Proof.
move=>C; case: (C)=>Vh [s1][s2][E]C1 C2; case:C1=>Vh' [hS'][hO']Z; subst s1.
move/cohV: (C)=>V; case/(scancel V): E; case=> Z1 Z2 _ Z3; subst hS' hO' s2.
by case: (coh_is_lock C2); do![case].
Qed.

Lemma is_alloc_stable : menv_stable Alloc (is_lock Alloc tt tt). 
Proof. by move=>s1 s2 L1 /(is_lock_stable L1); case; case. Qed.

(* try-alloc procedure - it might fail, hence the result is (Option ptr)! *)
Program Definition try_alloc : 
  {hS : heap}, STsep [W] (fun i => exists st (hO: heap),
                          i = hp ->> [hS, tt, hO] \+ st,
                  fun v m =>  
                      if v is Some p  
                      then exists st' (hO': heap) B (b: B), 
                           m = hp ->> [p :-> b \+ hS, tt, hO'] \+ st'
                      else exists st' (hO': heap), 
                           m = hp ->> [hS, tt, hO'] \+ st') :=
  Do (lock;;
      avl <-- [priv_read hp _ sent];
      if avl is p::ps
      then [priv_write hp sent ps];;
           unlock pfF;;
           ret (Some p)
      else unlock pfF;;
           ret None).
Next Obligation. by apply: entangle_extL. Defined.
Next Obligation. by apply: entangle_extL. Defined.
Next Obligation.
apply: gh=>i1 hS /=[st1][hO1] Ei1 A1 C1; subst i1.
apply: step=>//.  
apply: val_do'=>[|? i2]. 
- exists hS, hO1, tt, tt, st1; split=>//; last by move:(coh_is_lock C1). 
case/(_ hS hO1 tt tt st1 erefli (coh_is_lock C1))=>h'; case; case; case=>hO2 [st2][Ei2]A2 _ Inv C2.
case:Inv=>avl[h][Vh]Eh Hd. subst h'; rewrite (joinC hS) -joinA in Ei2; subst i2.
apply:step; apply: val_extend; first by move/split_cohP: C2.
apply: (gh_ex avl); apply: (gh_ex (h \+ hS)). 
apply: val_do=>[_|v i3]; first by exists hO2.
case=> Z1 [hO3] Ei3 Cp3 st3 C3 Mi3; subst i3 v.
case: avl Hd Vh C2 Cp3 C3 =>[Hd Vh C2 Cp3 C3|p ps Hd Vh C2 Cp3 C3].
- have Z: h = Unit by move: Hd. subst h; rewrite unitL.
  apply:step; apply: val_do'.
  + exists tt, tt, hS, hO3, (sent :-> ([::]:seq ptr)), st3.
    rewrite (joinC hS); split=>//. 
    by exists [::], Unit; rewrite unitR; split=>//; move/validL: Vh.
    by move: (is_locked_stable A2 Mi3).
  case=>i4; case/(_ tt tt hS hO3 (sent :-> ([::]:seq ptr)) st3).
  + rewrite (joinC hS); split=>//. 
    by exists [::], Unit; rewrite unitR; split=>//; move/validL: Vh.
    by move: (is_locked_stable A2 Mi3).
  move=>hO4; case; case=>[[st4] [Ei4] L4] C4; subst i4; apply:val_ret=>i5 M5.  
  by case:(menvs_stable is_alloc_stable M5 L4)=>hO5[st5][->]A5; exists st5, hO5. 
apply:step; apply: val_do'.
- exists (hp ->> [sent :-> (p :: ps) \+ (h \+ hS), tt, hO3]), st3; do![split=>//].
  by exists (h \+ hS), hO3, (seq ptr), (p::ps).
case=>i4; move/(_ (hp ->> [sent :-> (p :: ps) \+ (h \+ hS), tt, hO3]) st3 erefli Cp3).
case=>i5[st4][->{i4}]/=. move/(_ (h \+ hS)); case; first by exists hO3, (seq ptr), (p::ps).
move=>_[hO4]Ei5; subst i5=>M4 C4.
move: Hd; simpl=>[[h']][Z]I2; rewrite joinA; apply: step.
have I3: inv_of ((F tt \+ tt): pcm_of alloc_lock_info) (sent :-> ps \+ h').
exists ps, h'. split=>//.
  - rewrite Z (joinC (p:->_)) joinA in Vh; move/validL: Vh=>Vh.
    by rewrite 2!hvalidPtUn in Vh *. 
apply: val_do'. 
- exists tt, tt, (p :-> tt \+ hS), hO4, (sent :-> ps \+ h'), st4.
  rewrite {1}Z  -(joinCA (sent :-> _)) -!joinA (joinC hS); 
  split; first by []; first by []; last by rewrite !joinA.
  move:(is_locked_stable A2 Mi3)=>A3. 
  by move:(is_locked_stable A3 M4).
case=>i4 /(_ tt tt (p :-> tt \+ hS) hO4 (sent :-> ps \+ h') st4).
case. split; first by []; first by [].
- move:(is_locked_stable A2 Mi3)=>A3. 
  by move:(is_locked_stable A3 M4).
- by rewrite Z -(joinCA (sent :-> _)) -!joinA (joinC hS). 
move=>hO5; case; case;case=>[st5][Ei5 A5] C5; subst i4.
apply:val_ret=>i5 M5. 
case:(menvs_stable is_alloc_stable M5 A5)=>hO6 [st6][-> A6].
by exists st6, hO6, unit, tt.
Qed.

(********************************************)
(****** Main allocator procedure ************)
(********************************************)

Definition alloc_tp :=
  {hS : heap}, STsep [W] 
    (fun i => exists st (hO : heap), i = hp ->> [hS, tt, hO] \+ st,
     fun p m => exists st' (hO' : heap) B (b : B), 
                  m = hp ->> [p :-> b \+ hS, tt, hO'] \+ st').

(* Allocator runs in the loop until allocation is successful *)
Program Definition alloc_proc := 
  ffix (fun (loop : unit -> alloc_tp) xx =>
    Do (res <-- try_alloc; 
        if res is Some p 
        then ret p
        else loop tt)) tt.
Next Obligation.
apply:gh=>i1 hS [st][hO] Ei1 C1; subst i1.
apply: step; apply (gh_ex hS); apply: val_do=>//=[?|v i2]; first by exists st, hO. 
case: v=>//[p[st'][hO'][B][b] Ei2 C2|[st'][hO'] Ei2 C2]; subst i2.
- apply:val_ret=>i3 Mi2. 
  case:(menvs_stable is_alloc_stable Mi2 (coh_is_lock C2))=>{Mi2}hO2 [st2][->{i3}]A2.
  by exists st2, hO2, B, b. 
apply:(gh_ex hS); apply:val_do=>[_|p i3[st2][hO2][B][b]E3 C3]; first by exists st', hO'.
by exists st2, hO2, B, b. 
Qed.

Definition free_tp (p : ptr) :=
  {hS : heap}, STsep [W] 
                 (fun i => exists st (hO: heap) B (b: B),
                      i = hp ->> [p :-> b \+ hS, tt, hO] \+ st,
                  fun (_ : unit) m => exists st' (hO' : heap) B (b : B), 
                      m = hp ->> [hS, tt, hO'] \+ st').

Program Definition free_proc p : free_tp p := 
    Do ([priv_write hp p tt] ;;
        lock;;
        avl <-- [priv_read hp _ sent];
        [priv_write hp sent (p::avl)];;
        unlock pfF).
Next Obligation. by apply: entangle_extL. Defined.
Next Obligation. by apply: entangle_extL. Defined.
Next Obligation. by apply: entangle_extL. Defined.
Next Obligation.
apply: gh=>i1 hS [st1][hO1][B][b]->{i1} C1.
apply:step; apply:val_extend; first by  move/split_cohP: C1.
apply: (gh_ex hS); apply:val_do; first by exists hO1, B, b.
case=>i2 [_][hO2]->{i2}Cp2 st2 C2 M2.
move:(is_alloc_stable (coh_is_lock C1) M2)=> A2.
apply: step; apply: val_do'=>[|? i3].
- exists (p :-> tt \+ hS), hO2, tt, tt, st2; split=>//.
case/(_ (p :-> tt \+ hS) hO2 tt tt st2)=>//.
move=>hA; case; case=>[[hO3][st3][->{i3}]A3 _ Inv C3].
case: Inv=>avl [hA'][Va]Ea Ek; subst hA.
rewrite (joinC (p:->_ \+ _)) -!joinA in C3 *.
apply: step; apply:val_extend; first by apply/split_cohP: C3.
apply: (gh_ex avl); apply: (gh_ex (hA' \+ (p :-> tt \+ hS))).
apply: val_do; first by  exists hO3.
move=>x i4[->{x}][hO4]->{i4}Cp4 st4 C4 M4.
move:(is_alloc_stable (coh_is_lock C3) M4)=>A4.
apply: step; apply: val_extend=>//; apply:(gh_ex (hA' \+ (p :-> tt \+ hS))).
apply: val_do; first by exists hO4, (seq ptr), avl.
case=>// i5[_][hO5]->{i5}Cp5 st5 C5 M5.
rewrite -(joinC hS) 2!joinA -(joinC hS) -2!joinA (joinC hA') in Cp5 C5 *.
have I2: inv_of ((F tt \+ tt): pcm_of alloc_lock_info) 
                (sent :-> (p :: avl) \+ (p :-> tt \+ hA')).
  exists (p::avl), (p:->tt \+ hA'); split=>//; last by exists hA'. 
  by move/cohH:Cp5; rewrite /validH priv_flatsE; move/validL/validR.
apply: val_do'. 
- exists tt, tt, hS, hO5, (sent :-> (p :: avl) \+ (p :-> tt \+ hA')), st5.
  split=>//; last by move:(is_locked_stable (is_locked_stable A3 M4) M5).
case=>i6 /(_ tt tt hS hO5 (sent :-> (p :: avl) \+ (p :-> tt \+ hA')) st5);case=>//. 
split=>//; last by move:(is_locked_stable (is_locked_stable A3 M4) M5).
move=>hO6; case; case;case=>[st6][->{i6}]A6 C6.
by exists st6, hO6, unit, tt. 
Qed.

End DummyAllocator.

(* Setting up for exports; but these definitions shouldn't all be exported *)
Section CASAllocExport.
Require Import caslock.
Variables (pv al : nat) (sent lk : ptr).

Definition lkT := LockT (alloc_lock_info al sent) lk. 
Definition AllocCAS := cas_lockmod lkT.

(***************************************************)
(**************** Alloc procedure ******************)
(***************************************************)


(* now we give specs to alloc and dealloc in terms of getters of priv *)
(* we don't expose the getters of alloc as we don't plan to reason about *)
(* the internals of the allocator for now *)
Program Definition alloc : 
  {h : heap}, STsep [entangle (Priv pv) AllocCAS] 
                (fun i => self (pvg pv) i = h, 
                 fun p m => exists B (v : B), self (pvg pv) m = p :-> v \+ h) :=
  Do (alloc_proc pv AllocCAS).
Next Obligation.
apply: gh=>i h P C; case: C (C) P=>_ [xp][xa][->{i}] C1 C2 C <-.
apply: val_do'.
- exists (self (pvg pv) xp), xa, (other (pvg pv) xp). 
  by rewrite {1}(cohE C1) pv_cohJ.
move=>x m /(_ (self (pvg pv) xp)) XX.
have: exists st hO, xp \+ xa = pv ->> [self (pvg pv) xp, tt, hO] \+ st.
- by exists xa, (other (pvg pv) xp); rewrite {1}(cohE C1) pv_cohJ.
case/XX=>xa1 [hO'][B][v] ->{m XX} /cohV V'.
by exists B, v; rewrite getV // (getC C1 C).
Qed.

(***************************************************)
(***************** Free procedure ******************)
(***************************************************)

Program Definition dealloc p : 
  {h : heap}, STsep [entangle (Priv pv) AllocCAS] 
                (fun i => exists B (v : B), self (pvg pv) i = p :-> v \+ h,
                 fun (_ : unit) m => self (pvg pv) m = h) := 
  Do (free_proc pv AllocCAS p).
Next Obligation.
apply: gh=>i h [B][v] P C; case: C (C) P=>_ [xp][xa][->{i}] C1 C2 C P.
rewrite (getC C1 C) in P; apply: val_do'.
- exists h, xa, (other (pvg pv) xp), B, v.
  by rewrite {1}(cohE C1) /= P pv_cohJ.
case=>m /(_ h) /= XX. 
have: exists st hO B (v : B), xp \+ xa = pv ->> [p :-> v \+ h, tt, hO] \+ st.
- exists xa, (other (pvg pv) xp), B, v. 
  by rewrite {1}(cohE C1) /= pv_cohJ P.
by case/XX=>xa1 [hO'][B'][v'] ->{m XX} /cohV V'; rewrite getV.
Qed.

End CASAllocExport.

Module Exports.
Section Exports. 
Require Import caslock.
Variables (pv al : nat) (sent lk : ptr).

Definition AllocCAS : mod := AllocCAS al sent lk.
Notation al_coh := (Mod.coh AllocCAS).

Definition alg := 
  Getter al [encoded_pcm of lift mutexUnlifted * unit] (Unit : heap).

Lemma al_cohE s : 
        s \In al_coh -> s = al ->> [self alg s, joint alg s, other alg s].
Proof. by case=>_ _ [mL][mE][gL][gE][b][h][/= E _]; rewrite {-1}E !getE. Qed.

Definition allocGCMixin := GCMixin al_cohE.
Canonical al_getcoh := Eval hnf in GC alg al_coh allocGCMixin.

Lemma al_cohJ s : s \In al_coh -> valid (joint alg s).
Proof. 
case=>Vh Vs [mL][mE][gL][gE][b][h][/= E _].
by move: Vh; rewrite /validH E getE (@lock_flatsE (lkT al sent lk)). 
Qed.

Definition alloc : 
 {h : heap}, STsep [entangle (Priv pv) AllocCAS] 
   (fun i => self (pvg pv) i = h, 
    fun p m => exists B (v : B), self (pvg pv) m = p :-> v \+ h) := 
  (alloc pv al sent lk).

Definition dealloc p : 
  {h : heap}, STsep [entangle (Priv pv) AllocCAS] 
                (fun i => exists B (v : B), self (pvg pv) i = p :-> v \+ h,
                 fun (_ : unit) m => self (pvg pv) m = h) := 
  (dealloc pv al sent lk p).

End Exports. 

Notation al_coh al sent lk := (Mod.coh (AllocCAS al sent lk)).

End Exports. 

(* Ticketed lock allocator, just for fun, but we don't export it *)
Module TicketedLockAllocator.
Require Import ticketed.
Section TicketedLockAllocator.
Variables (hp al : nat) (sent owner next: ptr).

Definition lkT := TLockT owner next (alloc_lock_info al sent).
Definition AllocTk := ticketed_lockmod lkT.

End TicketedLockAllocator.
End TicketedLockAllocator.

End DummyAllocator.

Export DummyAllocator.Exports.


