Require Import ssreflect ssrbool ssrnat eqtype ssrfun seq path.
Require Import pred prelude idynamic ordtype finmap pcm coding unionmap heap.
Require Import feap state zigzag world inject rely hide atomic autosplit.
Require Import privstate footprint model lemmas stsep histories.
Set Implicit Arguments.
Unset Strict Implicit.
Unset Printing Implicit Defensive.

(*******************************************)
(********* Atomic pair snapshot ************)
(*******************************************)

Structure readpairT := ReadPairT {
  label_of : nat;
  eset_of : encoded_set;
  x_of : ptr; 
  y_of : ptr}.

Section ReadpairDefs.
Variable rT : readpairT.
Notation A := (eset_of rT). 
Notation st := [encoded_set of nat * A * A].
Notation st2 := [encoded_set of st * st].

Implicit Type h : hist st.

Definition stamps_mono h :=
  forall (T1 T2 : nat) (txy1 txy2 : st), 
    [txy1 in h at T1] -> [txy2 in h at T2] -> T1 <= T2 -> txy1.1.1 <= txy2.1.1.


Definition stamps_inj h :=
  forall t x1 x2 y1 y2, [(t, x1, y1) in h] -> [(t, x2, y2) in h] -> x1 = x2. 

Definition stamps_bound h tx := forall t x y, [(t, x, y) in h] -> t <= tx.

End ReadpairDefs.


Module ReadpairCoh.
Section ReadpairCoh.
Variable rT : readpairT.
Notation A := (eset_of rT).
Notation st := [encoded_set of nat * A * A].
Notation st2 := [encoded_set of st * st].
Notation l := (label_of rT).
Notation x := (x_of rT).
Notation y := (y_of rT).

(* The coherence predicate binds the actual contents of the two
   pointers, x and y to the last entry in the joinst history. It also
   makes sure that the history is well-formed and a subject of several
   properties, such as stamps_mono and stamps_inj. *)

Definition coh := 
  [Pred s | exists (hs ho : hist st) (tvx tvy : nat * A), 
    [/\ s = l ->> [hs, x :-> tvx \+ y :-> tvy, ho], 
        validH s,
        [(tvx, tvy.2) in hs \+ ho at last_key (hs \+ ho)],
        stamps_mono (hs \+ ho) /\ stamps_inj (hs \+ ho) /\ 
        stamps_bound (hs \+ ho) tvx.1 &
        well_formed (hs \+ ho)]].

Lemma readpair_flatsE (hs ho : hist st) (h : heap) :
        flats (l ->> [hs, h, ho]) = h.
Proof. by rewrite fsE flatsE !pcm_flatPt tp_flatPtE unitL unitR. Qed.

Section GetSO.
Notation sort := (to_type \o pcm_sort).
Notation V := (@idynamic pcm_code sort). 
Notation idyn_tp := (@idyn_tp _ sort). 

Lemma getSO_pf v : 
        idyn_tp v = pcm_natmap (decode st2) ->
        to_type (pcm_sort (idyn_tp v)) = hist st. 
Proof.
move=>-> /=; rewrite (_ : st_sort (decode A) = A) //. 
by move: (encdecE A); rewrite /encode. 
Qed.

Definition getSO (f : feap dpcm) : hist st := 
  if flook l f is Some v then 
    if @decP _ (idyn_tp v == pcm_natmap (decode st2)) idP is left pf 
    then icoerce id (idyn_val v) (getSO_pf (eqP pf)) else Unit
  else Unit.

Lemma getSOE h : getSO (l \-> h) = h.
Proof.
rewrite /getSO flookU eq_refl /=; case: decP=>// pf. 
move: (getSO_pf _) (esym _); rewrite /encode /=. 
have ->: st_sort (decode A) = A := encdecE A.  
by move=>H1 H2; rewrite !ieqc.
Qed.

Lemma getSOZ p1 p2 : 
        valid (p1 \* p2) -> 
        getSO (p1 \* p2) = getSO p1 \+ getSO p2.
Proof.
rewrite /getSO /= => V. 
case E: (flook l (p1 \* p2))=>[a|]; last first.
- move: (flook_none E) (flook_none E)=>{E}.
  rewrite {1}fdomZL fdomZR V.
  by do 2![case: fdom_flook=>// -> /=]; rewrite unitR.
rewrite (pcm_eta a) in E; case/flookZf: E=>{V} u1 [u2][E1] ->->.
case: decP=>/= H; last by rewrite unitR. 
rewrite /pcm_enc/code_enc/encode/pcm_val/code_val /= in u1 u2 E1 *.
rewrite {}E1; move: (esym _)=>E; rewrite !ieqc {E}.
do 3![move: (getSO_pf _)]; move: u1 u2. 
rewrite /to_type /decode {H}(eqP H) /=.
have -> : st_sort (decode A) = A := encdecE A. 
by move=>*; rewrite !ieqc.
Qed.

Definition self s : hist st := getSO (loc s).
Definition other s : hist st := getSO (env s). 

End GetSO.

(* Now functions and lemmas for extacting joint *)

Section GetJoint.
Notation sort := (to_type \o tp_sort).
Notation V := (@idynamic tp_code sort).
Notation idyn_tp := (@idyn_tp _ sort).

Lemma get_joint_pf v : 
        idyn_tp v = tp_heap ->
        to_type (tp_sort (idyn_tp v)) = heap.
Proof. by move=>->. Qed.

(* the next function extracts the joint heap *)
Definition joint s : heap := 
  if flook l (shr s) is Some v then 
    if @decP _ (idyn_tp v == tp_heap) idP is left pf 
    then icoerce id (idyn_val v) (get_joint_pf (eqP pf)) else Unit
  else Unit.

Lemma shr_joint s : s \In coh -> shr s = l ::-> joint s.
Proof.
by case=>hs [ho][tvx][tvy][->]; rewrite /joint fsE flookU eq_refl /= !ieqc.
Qed.

(* now we want to extract the two individual pointers *)
(* out of the joint heap, so we need some helper lemmas *)
(* to do dependent matching *)

Lemma xsome s v : 
        find x (joint s) = Some v -> s \In coh ->
        idynamic.idyn_tp v = prod nat A.
Proof.
move=>pf1 pf2; case: pf2 pf1=>hs [ho][tvx][tvy][->] Vh _ _ _.
rewrite /validH readpair_flatsE in Vh.
by rewrite fsE /joint flookU eq_refl /= !ieqc hfindPtUn //; case=><-. 
Qed.

Lemma ysome s v : 
        find y (joint s) = Some v -> s \In coh ->
        idynamic.idyn_tp v = prod nat A. 
Proof.
move=>pf1 pf2; case: pf2 pf1=>hs [ho][tvx][tvy][->] Vh _ _ _.
rewrite /validH readpair_flatsE joinC in Vh.
by rewrite fsE /joint flookU eq_refl /= !ieqc joinC hfindPtUn //; case=><-. 
Qed.

Lemma xnone s : find x (joint s) = None -> s \Notin coh.
Proof.
move=>pf1 pf2; case: pf2 pf1=>hs [ho][tvx][tvy][->] Vh _ _ _.
rewrite /validH readpair_flatsE in Vh.
by rewrite fsE /joint flookU eq_refl /= !ieqc hfindPtUn.
Qed.

Lemma ynone s : find y (joint s) = None -> s \Notin coh. 
Proof.
move=>pf1 pf2; case: pf2 pf1=>hs [ho][tvx][tvy][->] Vh _ _ _.
rewrite /validH readpair_flatsE joinC in Vh.
by rewrite fsE /joint flookU eq_refl /= !ieqc joinC hfindPtUn.
Qed.

(* now two functions for reading from x and y *)
Definition getx s : s \In coh -> nat * A := 
  match find x (joint s) as v' return _ = v' -> _ with
    Some v => fun pf1 pf2 => icoerce id (idyn_val v) (xsome pf1 pf2) 
  | None => fun pf1 pf2 => @False_rect _ (xnone pf1 pf2)
  end (erefl _).

Definition gety s : s \In coh -> nat * A := 
  match find y (joint s) as v' return _ = v' -> _ with
    Some v => fun pf1 pf2 => icoerce id (idyn_val v) (ysome pf1 pf2) 
  | None => fun pf1 pf2 => @False_rect _ (ynone pf1 pf2)
  end (erefl _).

Notation "'[getx'  pf ']'" := (@getx _ pf).
Notation "'[gety'  pf ']'" := (@gety _ pf).

(* the main property of getx and gety is that they *)
(* read the contents of x and y *)

Lemma get_joint s (pf : s \In coh) : 
        joint s = x :-> [getx pf] \+ y :-> [gety pf].
Proof.
move: (pf); case: pf=>hs [ho][tvx][tvy][E1] Vh _ _ _ pf.
rewrite {1}E1 /validH readpair_flatsE in Vh.
have E2: joint s = x :-> tvx \+ y :-> tvy.
- by rewrite /joint {1}E1 fsE flookU eq_refl /= !ieqc. 
rewrite /getx /gety.
case X: (find x (joint s)) (@xsome s) (@xnone s)=>[xb|] Xs; last by elim.
case Y: (find y (joint s)) (@ysome s) (@ynone s)=>[yb|] Ys; last by elim.
move: (Xs _ _ _) (Ys _ _ _)=>{Xs Ys} Xs Ys _ _.
case: xb Xs X yb Ys Y=>/= xT xb xE X [/= yT yb] yE Y; subst xT yT.
rewrite !ieqc.
move: X Y; rewrite E2 hfindPtUn // joinC.
rewrite joinC in Vh *; rewrite hfindPtUn // joinC.
by do 2![case; move/inj_pair2=><-].
Qed.

End GetJoint.

Notation "'[getx'  pf ']'" := (@getx _ pf).
Notation "'[gety'  pf ']'" := (@gety _ pf).

(* Now the main view which allows us to remove the existentials *)
(* when destructing the coh predicate *)

Lemma coh_view s (pf : s \In coh) : 
        let: tvx := [getx pf] in
        let: h := self s \+ other s in
        [/\ s = l ->> [self s, joint s, other s], validH s,
            [(tvx, [gety pf].2) in h at last_key h],
            stamps_mono h /\ stamps_inj h /\ stamps_bound h tvx.1 &
            well_formed h].
Proof.
move: (pf); case: pf=>hs [ho][tvx][tvy][/= E1] Vh H1 H2 H3 pf.
have [hsE hoE] : self s = hs /\ other s = ho.
- by rewrite E1 /self /other fsE !getSOE. 
have E2 : joint s = x :-> tvx \+ y :-> tvy.
- move: (shr_joint pf); rewrite {1}E1 fsE /=.
  by move/(fcancelPt (C:=typeCode)). 
suff [xE yE] : [getx pf] = tvx /\ [gety pf] = tvy.
- by rewrite E2 hsE hoE xE yE.
move: (get_joint pf); rewrite E2.
rewrite {1}E1 /validH readpair_flatsE in Vh.
by case/(hcancelV Vh)=>-> {Vh} Vh /(hcancelPtV Vh). 
Qed.

Lemma coh_viewE s : s \In coh -> s = l ->> [self s, joint s, other s].
Proof. by move=>C; case: (coh_view C). Qed.

Lemma coh_view' s : s \In coh <-> 
        [/\ s = l ->> [self s, joint s, other s], validH s &
            exists pf : s \In coh, 
              let: tvx := [getx pf] in
              let: h := self s \+ other s in
                [/\ [(tvx, [gety pf].2) in h at last_key h],
                    stamps_mono h, stamps_inj h, stamps_bound h tvx.1 &
                    well_formed h]].
Proof.
split.
- by move=>C; case: (coh_view C)=>E ?? [? [??]] ?; split=>//; exists C.
case=>E Vh [C][H1 H2 H3 H4 H5].
by exists (self s), (other s), [getx C], [gety C]; rewrite -get_joint.
Qed.

(* a few helper zagziging properties *)

Lemma zig_self s p : validSO (s \< p) -> self (s \< p) = self s \+ getSO p.
Proof. 
rewrite /validSO /self /other !spP; move/allvalidV. 
by rewrite validZE; case/andP=>V _; rewrite getSOZ // joinC.
Qed.

Lemma zag_self s p : self (s \> p) = self s.
Proof. by rewrite /self spP. Qed.

Lemma zig_other s p : validSO (s \< p) -> other (s \> p) = getSO p \+ other s.
Proof.
rewrite /validSO /self /other !spP; move/allvalidV.
by rewrite zipAC validZE zipC; case/and3P=>V _ _; rewrite getSOZ // joinC.
Qed.

Lemma zag_other s p : other (s \< p) = other s.
Proof. by rewrite /other spP. Qed.

Definition r2P := 
  (zig_self, zag_self, zig_other, zag_other).

(********************************)
(* coh is a coherence predicate *)
(********************************)

Lemma cohSO s : s \In coh -> validSO s. 
Proof. 
move=>C; case: (coh_view C)=>E _ H _ _.
rewrite /validSO E !fsE zipPt allvalidPt.
by case: H=>s' /find_some /dom_valid. 
Qed.

Lemma cohH s : s \In coh -> validH s.
Proof. by case/coh_view'. Qed.

Lemma coh_zigzag s p : s \< p \In coh <-> s \> p \In coh.
Proof.
move: s p.
have zagzig_validE s p : 
       validSO (s \< p) ->
       self (s \< p) \+ other (s \< p) = self (s \> p) \+ other (s \> p).
- by move=>V; rewrite !r2P // joinA. 
have zagzigE s p (h : heap) : 
        s \< p = l ->> [self (s \< p), h, other (s \< p)] <->
        s \> p = l ->> [self (s \> p), h, other (s \> p)].
- rewrite /self /other /= !spP -[s]st_eta zigE zagE /= !fsE; split.
  - case/stE=>/= /zipPtf /= [u1][t2][H1 ->->] ->->.
    by rewrite zipPt -!fsE !getSOE. 
  case/stE=>/= ->-> /zipPtf [u1][u2][H1 ->->].
  by rewrite zipPt -!fsE !getSOE.
move=>s p; split=>C; case: (coh_view C)=>E Vh H1 H2 H3;
rewrite get_joint in E; move/zagzigE: E => E;
have Vso : validSO (s \< p)
  by [rewrite ?(cohSO C) // ?validSO_zigzag ?(cohSO C)].
- rewrite validH_zigzag // in Vh.
  rewrite zagzig_validE // in H1 H2 H3.
  by exists (self (s \> p)), (other (s \> p)), [getx C], [gety C].  
rewrite -validH_zigzag // in Vh.
rewrite -zagzig_validE // in H1 H2 H3.
by exists (self (s \< p)), (other (s \< p)), [getx C], [gety C].  
Qed.

Lemma cohfV s : s \In coh -> footV_eq s (mem_seq' [:: l]).
Proof. by move/coh_viewE=>->; rewrite fsE. Qed.

Lemma coh_weq s s' : s \In coh -> s' \In coh -> weq typeCode (shr s) (shr s').
Proof. by move/coh_viewE=>-> /coh_viewE ->; rewrite !fsE /=. Qed.

Lemma coh_sig s : s \In coh -> l \-> (Unit : hist st) = [zunit loc s].
Proof. by move/coh_viewE=>->; rewrite fsE zunitPt. Qed.

Definition ReadPairCohMixin :=
  CohPredMixin cohSO cohH cohfV coh_weq coh_sig coh_zigzag. 
Definition readpair_coh := Eval hnf in CohPred coh ReadPairCohMixin.

(* Additional zagzigging lemma that could be proved *)
(* once we know readpair_coh is a valid coherence predicate *)

Section GetZ.
Variables (s : state) (p : feap dpcm).
Variables (C1 : s \< p \In readpair_coh) (C2 : s \> p \In readpair_coh).

Lemma zagzig_joint : joint (s \< p) = joint (s \> p).
Proof. 
move/coh_zigzag: (C1)=>C1'; move: (shr_joint C1) (shr_joint C1').
by rewrite !spP=>-> /(fcancelPt (C:=typeCode)). 
Qed.

Lemma zagzig_get : [getx C1] = [getx C2] /\ [gety C1] = [gety C2].
Proof.
move: (cohH C1) zagzig_joint.
rewrite {1}(coh_viewE C1) /validH readpair_flatsE !get_joint => V.
by case/(hcancelV V)=>-> {V} V /(hcancelPtV V). 
Qed.

Lemma zagzig_getx : [getx C1] = [getx C2]. Proof. by case: zagzig_get. Qed.

Lemma zagzig_gety : [gety C1] = [gety C2]. Proof. by case: zagzig_get. Qed.

End GetZ.
End ReadpairCoh.


Module Exports.
Notation r2_coh := readpair_coh.
Notation r2_flatsE := readpair_flatsE.
Notation r2_getSO := getSO.
Notation r2_getSOE := getSOE.
Notation r2_getSOZ := getSOZ.
Notation r2_self := self.
Notation r2_other := other.
Notation r2_joint := joint.
Notation r2_shr_joint := shr_joint.
Notation "'[getx'  pf ']'" := (@getx _ _ pf).
Notation "'[gety'  pf ']'" := (@gety _ _ pf).
Notation r2_get_joint := get_joint.
Notation r2_coh_view := coh_view.
Notation r2_coh_viewE := coh_viewE.
Notation r2_coh_view' := coh_view'.

Definition r2P := 
  (zig_self, zag_self, zig_other, zag_other, zagzig_joint, zagzig_getx, zagzig_gety).

End Exports.

End ReadpairCoh.

Export ReadpairCoh.Exports.

(************************)
(* Readpair transitions *)
(************************)

Module WriteXTrans.
Section WriteXTrans.
Variable rT : readpairT.
Notation A := (eset_of rT).
Notation st := [encoded_set of nat * A * A].
Notation st2 := [encoded_set of st * st].
Notation l := (label_of rT). 
Notation x := (x_of rT). 
Notation y := (y_of rT).    
Notation coh := (r2_coh rT).
Notation self := (r2_self rT).
Notation other := (r2_other rT).
Notation joint := (r2_joint rT).
Notation getSO := (r2_getSO rT).

(* Write to x pointer atomically, augment self-history with a new entry *)
Definition writex_rel s s' :=
  exists (pf : s \In coh) (pf' : s' \In coh), 
    let: (tx, vx) := ([getx pf].1, [getx pf].2) in
    let: (vx', vy) := ([getx pf'].2, [gety pf].2) in
    let: h := self s \+ other s in
    [/\ self s' = self s \+ fresh h \\-> ((tx, vx, vy), (tx.+1, vx', vy)), 
        other s' = other s, [getx pf'].1 = tx.+1 & [gety pf'] = [gety pf]].

Lemma writex_coh s s' : writex_rel s s' -> s \In coh /\ s' \In coh.
Proof. by case=>C [C']. Qed.

Lemma writex_env s s' : writex_rel s s' -> env s = env s'.
Proof. 
by case=>C [C'][_ E _ _]; rewrite (r2_coh_viewE C) (r2_coh_viewE C') !fsE /= E.
Qed.

Lemma writex_zagzig s s' p : 
        env s = env s' ->
        writex_rel (s \> p) (s' \> p) -> writex_rel (s \< p) (s' \< p).
Proof.
move=>E [C][C'][sE oE tE gyE]; move/coh_zigzag: (C)=>D; move/coh_zigzag: (C')=>D'.
exists D, D'; split; last by [rewrite !r2P]; last by rewrite !r2P. 
- by rewrite !r2P ?(cohSO D) ?(cohSO D') // -!joinA in sE *; rewrite joinCA joinC sE.
by [rewrite !r2P /other E]. 
Qed.

Lemma writex_footH s s' : writex_rel s s' -> footeqH (flats s) (flats s').
Proof.
case=>C [C'] _; rewrite (r2_coh_viewE C) (r2_coh_viewE C').
by rewrite !r2_flatsE !r2_get_joint; apply: footeqHUn; apply: footeqHPt.
Qed.

End WriteXTrans.

Module Exports.

Notation writex_rel := writex_rel.

Definition WriteXGuarMixin rT := 
  GuarMixin (@writex_coh rT) (@writex_env rT) (@writex_zagzig rT). 
Canonical writex_guar rT := 
  Eval hnf in Guar (writex_rel rT) (WriteXGuarMixin rT).

Definition WriteXIntGuarMixin rT := IntGuarMixin (@writex_footH rT).
Canonical WriteXIntGuar rT := 
  Eval hnf in IntGuar (writex_guar rT) (WriteXIntGuarMixin rT).

End Exports.
End WriteXTrans.

Export WriteXTrans.Exports.


Module WriteYTrans.
Section WriteYTrans.
Variable rT : readpairT.
Notation A := (eset_of rT).
Notation st := [encoded_set of nat * A * A].
Notation st2 := [encoded_set of st * st].
Notation l := (label_of rT). 
Notation x := (x_of rT). 
Notation y := (y_of rT).    
Notation coh := (r2_coh rT).
Notation self := (r2_self rT).
Notation other := (r2_other rT).
Notation joint := (r2_joint rT).
Notation getSO := (r2_getSO rT).

Definition writey_rel s s' :=
  exists (pf : s \In coh) (pf' : s' \In coh), 
    let: (tx, vx) := ([getx pf].1, [getx pf].2) in
    let: (vy, vy') := ([gety pf].2, [gety pf'].2) in
    let: h := self s \+ other s in
    [/\ self s' = self s \+ fresh h \\-> ((tx, vx, vy), (tx, vx, vy')), 
        other s' = other s & [getx pf'] = [getx pf]].

Lemma writey_coh s s' : writey_rel s s' -> s \In coh /\ s' \In coh.
Proof. by case=>C [C']. Qed.

Lemma writey_env s s' : writey_rel s s' -> env s = env s'.
Proof. 
by case=>C [C'][_ E _]; rewrite (r2_coh_viewE C) (r2_coh_viewE C') !fsE /= E.
Qed.

Lemma writey_zagzig s s' p : 
        env s = env s' ->
        writey_rel (s \> p) (s' \> p) -> writey_rel (s \< p) (s' \< p).
Proof.
move=>E [C][C'][sE oE gyE]; move/coh_zigzag: (C)=>D; move/coh_zigzag: (C')=>D'.
exists D, D'; split; last by [rewrite !r2P]; last by [rewrite !r2P /other E]. 
rewrite !r2P ?(cohSO D) ?(cohSO D') // -!joinA in sE *.
by rewrite joinCA joinC sE.
Qed.

Lemma writey_footH s s' : writey_rel s s' -> footeqH (flats s) (flats s').
Proof.
case=>C [C'] _; rewrite (r2_coh_viewE C) (r2_coh_viewE C').
by rewrite !r2_flatsE !r2_get_joint; apply: footeqHUn; apply: footeqHPt.
Qed.

End WriteYTrans.

Module Exports.

Notation writey_rel := writey_rel.

Definition WriteYGuarMixin rT := 
  GuarMixin (@writey_coh rT) (@writey_env rT) (@writey_zagzig rT). 
Canonical writey_guar rT := 
  Eval hnf in Guar (writey_rel rT) (WriteYGuarMixin rT).

Definition WriteYIntGuarMixin rT := IntGuarMixin (@writey_footH rT).
Canonical WriteYIntGuar rT := 
  Eval hnf in IntGuar (writey_guar rT) (WriteYIntGuarMixin rT).

End Exports.
End WriteYTrans.

Export WriteYTrans.Exports.

(* Now build the whole concurroid *)

Canonical ReadPairCore rT : mod_core := 
  @Mod.Core (r2_coh rT) [PredU (Pred1 (WriteXIntGuar rT)) & 
                               (Pred1 (WriteYIntGuar rT))].
Canonical ReadPair rT : mod := @Mod.Make (ReadPairCore rT) Pred0. 

Module ReadXAct.
Section ReadXAct.
Variable rT : readpairT.
Notation A := (eset_of rT).
Notation st := [encoded_set of nat * A * A].
Notation st2 := [encoded_set of st * st].
Notation l := (label_of rT). 
Notation x := (x_of rT). 
Notation y := (y_of rT).    
Notation coh := (r2_coh rT).
Notation self := (r2_self rT).
Notation other := (r2_other rT).
Notation joint := (r2_joint rT).
Notation getSO := (r2_getSO rT).
Notation guar := (Mod.internal (ReadPair rT)). 

Definition readx_safe s := s \In coh.

Definition readx_step s s' v := 
  exists pf : s \In coh, v = [getx pf] /\ s' = s.
    
Lemma readx_safe_coh : safe_coh coh readx_safe.
Proof. by []. Qed.

Lemma readx_safe_zagzig : safe_zagzig readx_safe.
Proof. by move=>s p; rewrite coh_zigzag. Qed.

Lemma readx_step_safe : step_safe readx_safe readx_step.
Proof. by move=>s s' v; case. Qed.

Lemma readx_step_guar : step_guar guar readx_step.
Proof. by move=>s s' v [C][_] ->; left. Qed.

Lemma readx_step_total : step_total readx_safe readx_step.
Proof. by move=>s C; exists s, [getx C], C. Qed.

Lemma readx_step_frame : step_frame readx_safe readx_step.
Proof.
move=>s s' p v C[C'][->]{v}->{s'}; exists s; split=>//.
by exists C; split=>//; rewrite r2P.
Qed.

Definition ReadX_pre := 
  Action readx_safe_coh readx_safe_zagzig readx_step_safe readx_step_guar 
         readx_step_frame readx_step_total.

End ReadXAct.

Lemma readx_step_erase rT : int_step_erase (ReadX_pre rT).
Proof.
move=>s1 h1 t1 v1 s2 h2 t2 v2.
move=>V E H1 H2; case: H1 H2 V E.
move=>C1[->]{v1}->{t1}.
case=>C2[->{v2}]->{t2} V E; split=>//.
move/r2_coh_viewE :(C2)=>T1;
move/r2_coh_viewE :(C1)=>T2.
rewrite T1 T2 !r2_flatsE !r2_get_joint -!joinA /= in V E.
by case/(hcancelV V) : E=>->.
Qed.

Module Exports.
Definition ReadX rT := IntAction (@readx_step_erase rT). 
End Exports.
End ReadXAct.

Export ReadXAct.Exports.

Module ReadYAct.
Section ReadYAct.
Variable rT : readpairT.
Notation A := (eset_of rT).
Notation st := [encoded_set of nat * A * A].
Notation st2 := [encoded_set of st * st].
Notation l := (label_of rT). 
Notation x := (x_of rT). 
Notation y := (y_of rT).    
Notation coh := (r2_coh rT).
Notation self := (r2_self rT).
Notation other := (r2_other rT).
Notation joint := (r2_joint rT).
Notation getSO := (r2_getSO rT).
Notation guar := (Mod.internal (ReadPair rT)). 

Definition ready_safe s := s \In coh.

Definition ready_step s s' v := 
  exists pf : s \In coh, v = [gety pf] /\ s' = s.
    
Lemma ready_safe_coh : safe_coh coh ready_safe.
Proof. by []. Qed.

Lemma ready_safe_zagzig : safe_zagzig ready_safe.
Proof. by move=>s p; rewrite coh_zigzag. Qed.

Lemma ready_step_safe : step_safe ready_safe ready_step.
Proof. by move=>s s' v; case. Qed.

Lemma ready_step_guar : step_guar guar ready_step.
Proof. by move=>s s' v [C][_] ->; left. Qed.

Lemma ready_step_total : step_total ready_safe ready_step.
Proof. by move=>s C; exists s, [gety C], C. Qed.

Lemma ready_step_frame : step_frame ready_safe ready_step.
Proof.
move=>s s' p v C[C'][->]{v}->{s'}; exists s; split=>//.
by exists C; split=>//; rewrite r2P.
Qed.

Definition Ready_pre := 
  Action ready_safe_coh ready_safe_zagzig ready_step_safe ready_step_guar 
         ready_step_frame ready_step_total.

End ReadYAct.

Lemma ready_step_erase rT : int_step_erase (Ready_pre rT).
Proof.
move=>s1 h1 t1 v1 s2 h2 t2 v2.
move=>V E H1 H2; case: H1 H2 V E.
move=>C1[->]{v1}->{t1}.
case=>C2[->{v2}]->{t2} V E; split=>//.
move/r2_coh_viewE :(C2)=>T1;
move/r2_coh_viewE :(C1)=>T2.
rewrite T1 T2 !r2_flatsE !r2_get_joint -!joinA in V E. 
move/validR: (V)=>V1; case:(hcancel V E)=> _ _ E1.
by case/(hcancelV V1):(E1)=>->.
Qed.

Module Exports.
Definition ReadY rT := IntAction (@ready_step_erase rT). 
End Exports.
End ReadYAct.

Export ReadYAct.Exports.


Section ReadPairProgram.
Variable rT : readpairT.
Notation A := (eset_of rT).
Notation st := [set of nat * A * A].
Notation st2 := [encoded_set of st * st].
Notation l := (label_of rT). 
Notation x := (x_of rT). 
Notation y := (y_of rT).    
Notation coh := (r2_coh rT).
Notation self := (r2_self rT).
Notation other := (r2_other rT).
Notation joint := (r2_joint rT).
Notation getSO := (r2_getSO rT).
Notation W := (ReadPair rT).

Definition hist_preord (h1 h2 : hist st) := 
  valid h1 = valid h2 /\ exists h, h2 = h1 \+ h.

Notation "'[' h1 '<<=' h2 ']'" := (hist_preord h1 h2).

Lemma hist_refl h : [h <<= h].
Proof. by split=>//; exists Unit; rewrite unitR. Qed.

Hint Resolve hist_refl.

Lemma valid_hist i : i \In coh -> valid (self i \+ other i).
Proof.
move=>C; move: (cohSO C).
by rewrite {1}(r2_coh_viewE C) /validSO !fsE /= zipPt allvalidPt. 
Qed.

(* Lemma for constructing existential t and vy *)
Lemma readxy_snap j (Cj : j \In coh) : 
        let: h := self j \+ other j in
        [([getx Cj], [gety Cj].2) in h at last_key h].
Proof. by case: (r2_coh_view Cj). Qed.

(* Monotonicity of histories wrt to Rely *)
Lemma r2_mrely s1 s2 : mrely W s1 s2 -> [other s1 <<= other s2].
Proof.
move=>H; case: (mrely_coh H)=>C1 C2; split.
- by rewrite (validR (valid_hist C1)) (validR (valid_hist C2)).
case: H; last by case=>ch [h][] /mem_chan [x][]. 
case=>[|[g]][]. 
- by move/transp_inj=>->; exists Unit; rewrite unitR. 
case=><-; case=>C [pf'][]; rewrite /self /other !loc_transp !env_transp;
by move=>->; eexists _.  
Qed.

Lemma hist_self s1 s2 : menv_steps W s1 s2 -> self s2 = self s1.
Proof. by rewrite /self; move/menvs_loc=>->. Qed.

Lemma hist_other s1 s2 : menv_steps W s1 s2 -> [other s1 <<= other s2].
Proof.
case=>n; elim: n s1 s2 => [|n IH] s1 s2 /=; first by case=>->. 
by case=>k [] /r2_mrely H /IH; apply: hist_trans H.
Qed.

Program Definition readx_proc : 
  {h}, STsep [W] (fun i => self i = Unit /\ [h <<= other i],
                  fun tvx m => exists t vy, 
                    [/\ last_key h <= t,
                        self m = Unit, [h <<= other m] &
                        [(tvx, vy) in other m at t]]) :=
  Do (act (ReadX rT)).
Next Obligation. 
apply: gh=>i h [Ei H] Ci; apply: val_act=>j Mi.
split=>[|_ _ m [Cj][->-> Mj]]; first by case/menvs_coh: Mi. 
exists (last_key (other j)), [gety Cj].2. 
move: (readxy_snap Cj)=>H'.
rewrite (hist_self Mj) (hist_self Mi) Ei unitL in H' *.
split=>//; last by apply: histE (hist_other Mj) H'. 
- by apply: histK (hist_trans H (hist_other Mi)).
by apply: hist_trans (hist_other (menvs_trans Mi Mj)).
Qed.


Program Definition ready_proc : 
  {h}, STsep [W] (fun i => self i = Unit /\ [h <<= other i],
                  fun tvy m => exists t tvx, 
                    [/\ last_key h <= t, 
                        self m = Unit, [h <<= other m] &
                        [(tvx, tvy.2) in other m at t]]) := 
  Do (act (ReadY rT)).
Next Obligation. 
apply: gh=>i h [Ei H] Ci; apply: val_act=>j Mi.
split=>[|_ _ m [Cj][->-> Mj]]; first by case/menvs_coh: Mi. 
exists (last_key (other j)), [getx Cj].
move: (readxy_snap Cj)=>H'.
rewrite (hist_self Mj) (hist_self Mi) Ei unitL in H' *.
split=>//; last by apply: histE (hist_other Mj) H'.
- by apply: histK (hist_trans H (hist_other Mi)).
by apply: hist_trans (hist_other (menvs_trans Mi Mj)).
Qed.

(***************************************************)
(************* ReadPair procedure ******************)
(***************************************************)

Definition read2_tp := 
  {h}, STsep [W] (fun i => self i = Unit /\ [h <<= other i],
                  fun vxy m => exists t tx, 
                    [/\ last_key h <= t, 
                        self m = Unit, [h <<= other m] &
                        [(tx, vxy.1, vxy.2) in other m at t]]).

Program Definition read_pair :=
  ffix (fun (loop : unit -> read2_tp) xx =>
    Do (tvx  <-- readx_proc; 
        tvy  <-- ready_proc;
        tmpx <-- readx_proc;
        if tmpx.1 == tvx.1 then ret (tvx.2, tvy.2)      
        else loop tt)) tt.
Next Obligation. 
apply: gh=>i1 h [S O] C {xx}. 
apply: step; apply: (gh_ex h); apply: val_do=>//.
case=>{i1 C S O} tx2 vx2 /= i2 [t2][vy2][N S O H2] C.
apply: step; apply: (gh_ex _); apply: val_do=>//.
case=>{C S} _ vy3 /= i3 [t3][[tx3 vx3]][N3 S O3 H3] C.
move/(leq_trans (last_key_in H2)): N3=>T23; move: (leq_trans N T23)=>{N} N.
move/(histE O3): H2=>H2; move/(hist_trans O): O3 => {i2 O} O.
apply: step; apply: (gh_ex _); apply: val_do=>//. 
case=>{C S} tx4 vx4 /= i4 [t4][vy4][N4 S O4 H4] C.
move/(leq_trans (last_key_in H3)): N4=>T34.
move/(histE O4): H2=>H2; move/(histE O4): H3=>H3.
move/(hist_trans O): O4 => {i3 O} O.
case: ifP=>E; last by apply: (gh_ex h); apply: val_do.
rewrite {tx4 E}(eqP E) in H4; apply: val_ret=>i5 M /=. 
exists t3, tx3; split=>//.
- by rewrite (hist_self M).
- by apply: hist_trans (hist_trans _ O) (hist_other M). 
apply: {M} histE (hist_other M) _.
have {vy4 H4} E : tx2 = tx3.
- have SM : stamps_mono (other i4).  
  - by case/r2_coh_view': C=>_ _ [?][_]; rewrite S unitL. 
  by apply: anti_leq; rewrite (SM _ _ _ _ H2 H3 T23) (SM _ _ _ _ H3 H4 T34).
rewrite {tx2}E in H2.
have {H2 vx2 vy2}-> // : vx2 = vx3.
have SI : stamps_inj (other i4).
- by case/r2_coh_view': C=>_ _ [X][_]; rewrite S unitL. 
by apply: SI (at_in H2) (at_in H3).
Qed.

End ReadPairProgram.










